/***********************************************************

mimeplugin.c

See CHANGES for a complete list of changes

************************************************************/


#include "mimeplugin.h"
#include <errno.h>


/**********************************************************************/

char*
getPluginPath( )
{
   char* pluginPath;

   enterFunction( "getPluginPath" );

   if ( ( pluginPath = getenv( "NPX_PLUGIN_PATH" ) ) == NULL ) {
      pluginPath = (char*)
            malloc( sizeof( char ) * ( 2 * MEMORY_CHUNK_SIZE ) );

      sprintf( pluginPath,
               "%s%s:%s/%s",
               NS_DIR,
               "plugins/",
               getenv( "HOME" ),
               ".netscape/plugins/" );
   }

   exitFunction( "getPluginPath" );

   return( pluginPath );
}  /* getPluginPath */

/**********************************************************************/


char*
getNextEntry( FILE* fptr,
              char* buff )
{
   enterFunction( "getNextEntry" );

   while ( fgets( buff, MAX_ENTRY_LENGTH, fptr ) != NULL ) {
      if ( ( buff[0] != '#' ) &&
           ( buff[0] != '\n' ) ) {
         if ( buff[strlen( buff ) - 1] == '\n' ) {
            buff[strlen( buff ) - 1] = '\0';
         }

         exitFunction( "getNextEntry" );

         return( buff );
      }  /* if line is not comment or NULL */
   }  /* while fgets != NULL */

   exitFunction( "getNextEntry" );

   return( NULL );
}  /* getNextEntry */


/**********************************************************************/

char*
removeLeadingSpaces( char* buff )
{
   enterFunction( "removeLeadingSpaces" );

   /* This should be safe */
   strcpy( buff,
           buff + strspn( buff, " " ) );

   exitFunction( "removeLeadingSpaces" );

   return( buff );
}  /* removeLeadingSpaces */


/**********************************************************************/

char*
removeTrailingSpaces( char* buff )
{
   int buffLen = strlen( buff );

   enterFunction( "removeTrailingSpaces" );

   while ( buff[ buffLen - 1 ] == ' ' ) {
      --buffLen;
   }

   buff[ buffLen ] = '\0';

   exitFunction( "removeTrailingSpaces" );

   return( buff );
}  /* removeTrailingSpaces */


/**********************************************************************/

char*
removeLeadingAndTrailingSpaces( char* buff )
{
   enterFunction( "removeLeadingAndTrailingSpaces" );

   removeLeadingSpaces( buff );
   removeTrailingSpaces( buff );

   exitFunction( "removeLeadingAndTrailingSpaces" );

   return( buff );
}  /* removeLeadingAndTrailingSpaces */


/**********************************************************************/

char*
removeAllSpaces( char* buff )
{
   char ch;
   char* newBuff;
   int i = 0,
       j = 0;

   enterFunction( "removeAllSpaces" );

   newBuff = (char*) malloc( sizeof( char ) * ( strlen( buff ) + 1 ) );

   while ( ( ch = buff[i++] ) != '\0' ) {
      if ( ch != ' ' ) {
         newBuff[j++] = ch;
      }
   }

   newBuff[j] = '\0';

   strcpy( buff,
           newBuff );

   free( newBuff );

   exitFunction( "removeAllSpaces" );

   return( buff );
}  /* removeAllSpaces */


/**********************************************************************/

char*
addPeriodToEachExtension( char* extensions )
{
   char extensionsCopy[MAX_FIELD_LENGTH];
   char* nextExtension;

   enterFunction( "addPeriodToEachExtension" );

   /* copy buff into the new buff */
   strcpy( extensionsCopy,
           extensions );

   /* Reset extensions to be empty */
   extensions[0] = '\0';

   /* get the first extension */
   nextExtension = (char*) strtok( extensionsCopy, ", " );

   while ( nextExtension != NULL ) {
      sprintf( extensions,
               "%s.%s,",
               extensions,
               nextExtension );

      nextExtension = (char*) strtok( NULL, ", " );
   }

   /* Remove trailing comma (',') */
   extensions[strlen( extensions ) - 1] = '\0';

   exitFunction( "addPeriodToEachExtension" );

   return( extensions );
}  /* addPeriodToEachExtension */


/**********************************************************************/

entryStruct*
splitFields( char* buff )
{
   entryStruct* entry = NULL;

   enterFunction( "splitFields" );

   /* Remove the new-line from the end */
   if ( buff[ strlen( buff ) - 1 ] == '\n' ) {
      buff[ strlen( buff ) - 1 ] = '\0';
   }  /* if the last character is a new-line */

   printDebug( "Working on: '%s'\n",
               buff );

   /* Make sure we have a valid entry line */
   if ( countNumberOfTokens( buff, ";" ) >= 3 ) {
      /* There are enough fields in the entry, we can continue */
      char* buffCopy;
      char* nextField;

      /* make a copy of buff */
      buffCopy = (char*) malloc( sizeof( char ) * ( strlen( buff ) + 1 ) );
      strcpy( buffCopy,
              buff );

      /* allocate space for the entry */
      entry = (entryStruct*) malloc( sizeof( entryStruct ) );

      /* Initialize everything to NULL */
      entry->mimeType = NULL;
      entry->extensions = NULL;
      entry->commandLine = NULL;
      entry->description = NULL;
      entry->next = NULL;

      /* get the MIME type */
      if ( ( nextField = (char*) strtok( buffCopy, ";" ) ) != NULL ) {

         /* Remove all extraneous spaces */
         removeLeadingAndTrailingSpaces( nextField );

         /* allocate space for and copy the MIME type */
         entry->mimeType = (char*)
               malloc( sizeof( char ) * ( strlen( nextField ) + 1 ) );
         strcpy( entry->mimeType,
                 nextField );
      }

      /* get the extensions */
      if ( ( nextField = (char*) strtok( NULL, ";" ) ) != NULL ) {
         /* Remove all extraneous spaces (happens to be all spaces) */
         removeAllSpaces( nextField );

         /* allocate space for and copy the extensions */
         entry->extensions = (char*)
               malloc( sizeof( char ) * ( strlen( nextField ) + 1 ) );
         strcpy( entry->extensions,
                 nextField );
      }

      /* get the command line (application and switches/options) */
      if ( ( nextField = (char*) strtok( NULL, ";" ) ) != NULL ) {
         /* Remove all extraneous spaces */
         removeLeadingAndTrailingSpaces( nextField );

         /* allocate space for and copy the commandLine */
         entry->commandLine = (char*)
               malloc( sizeof( char ) * ( strlen( nextField ) + 1 ) );
         strcpy( entry->commandLine,
                 nextField );
      }

      /* get the description */
      if ( ( nextField = (char*) strtok( NULL, ";" ) ) != NULL ) {
         /* Remove all extraneous spaces */
         removeLeadingAndTrailingSpaces( nextField );

         /* allocate space for and copy the description */
         entry->description = (char*)
               malloc( sizeof( char ) * ( strlen( nextField ) + 1 ) );
         strcpy( entry->description,
                 nextField );
      }

      if ( ( entry->mimeType == NULL ) ||
           ( entry->extensions == NULL ) ||
           ( entry->commandLine == NULL ) ) {
         /* something went wrong, deallocate the memory */
         freeEntryStruct( entry );
         entry = NULL;
      }

      /* free up the memory */
      free( buffCopy );
   }

   exitFunction( "splitFields" );

   return( entry );
}  /* splitFields */


/**********************************************************************/

int
countNumberOfTokens( char* str,
                     char* delimeters )
{
   int numberOfTokens = 0;

   enterFunction( "countNumberOfTokens" );

   if ( str != NULL ) {
      char* strCopy;
      char* nextToken;

      /* copy string */
      strCopy = (char*)
            malloc( sizeof( char ) * strlen( str ) + 1 );
      strcpy( strCopy,
              str );

      /* start to tokenize the string */
      nextToken = strtok( strCopy, delimeters );

      /* count the number of tokens */
      while ( (++numberOfTokens) &&
              (char*) strtok( NULL, delimeters ) != NULL ) {
      }

      /* free up the memory */
      free( strCopy );
   }

   exitFunction( "countNumberOfTokens" );

   return( numberOfTokens );
}  /* countNumberOfTokens */


/**********************************************************************/

char**
splitCommandLine( char* commandLine,
                  char* fileName )
{
   char* nextArgument;
   char* commandLineCopy;
   char** commandLineArray;
   int numberOfArguments = countNumberOfTokens( commandLine, " " );
   int i = 0;

   enterFunction( "splitCommandLine" );

   /* copy the commandLine */
   commandLineCopy = (char*)
         malloc( sizeof( char ) * ( strlen( commandLine ) + 1 ) );
   strcpy( commandLineCopy,
           commandLine );

   /* if there is not a '%s', we put the fileName at the end,
    * so we have to increment the numberOfArguments
    */
   if ( (char*) strstr( commandLine, "%s" ) == NULL ) {
      /* There is no '%s', we need to make the array one bigger */
      ++numberOfArguments;
   }

   /* allocate space for the commandLineArray (number of arguments +
    * NULL == numberOfArguments + 1)
    */
   commandLineArray = (char**)
         malloc( sizeof( char* ) * ( numberOfArguments + 1 ) );

   /* Terminate the array with NULL */
   commandLineArray[numberOfArguments] = NULL;

   /* Get the application name */
   nextArgument = (char*) strtok( commandLineCopy, " " );

   while ( nextArgument != NULL ) {
      /* allocate space for the argument and copy it */
      if ( strcmp( nextArgument, "%s" ) == 0 ) {
         /* Put the fileName here */
         commandLineArray[i] = (char*)
               malloc( sizeof( char ) * ( strlen( fileName ) + 1 ) );
         strcpy( commandLineArray[i],
                 fileName );
      }
      else {
         /* Put the argument here */
         commandLineArray[i] = (char*)
               malloc( sizeof( char ) * ( strlen( nextArgument ) + 1 ) );
         strcpy( commandLineArray[i],
                 nextArgument );
      }

      nextArgument = (char*) strtok( NULL, " " );

      /* Increment the array index */
      ++i;
   }  /* while there are more arguments */

   /* free the memory */
   free( commandLineCopy );

   /* if there is not a '%s', we put the fileName at the end */
   if ( (char*) strstr( commandLine, "%s" ) == NULL ) {
      commandLineArray[i] = (char*)
            malloc( sizeof( char ) * ( strlen( fileName ) + 1 ) );
      strcpy( commandLineArray[i],
              nextArgument );
      ++i;
   }

   if ( numberOfArguments != i ) {
      printError( "Number of arguments do not match\n" );
   }

   exitFunction( "splitCommandLine" );

   return( commandLineArray );
}  /* splitCommandLine */


/**********************************************************************/

void
freeEntryStruct( entryStruct* entry )
{
   enterFunction( "freeEntryStruct" );

   if ( entry != NULL ) {
      free( entry->mimeType );
      free( entry->extensions );
      free( entry->commandLine );
      free( entry->description );

      free( entry );
   }  /* if entry is real */

   exitFunction( "freeEntryStruct" );
}  /* freeEntryStruct */


/**********************************************************************/

entryStruct*
addOrReplaceEntry( entryStruct* head,
                   entryStruct* newEntry )
{
   entryStruct* current = head;
   entryStruct* previous = NULL;

   enterFunction( "addOrReplaceEntry" );

   while ( ( current != NULL ) &&
           ( strcmp( current->mimeType,
                     newEntry->mimeType ) < 0 ) ) {
      previous = current;
      current = current->next;
   }

   if ( current == NULL ) {
      /* insertion after current */
      newEntry->next = NULL;
   }
   else if ( strcmp( current->mimeType,
                     newEntry->mimeType ) == 0 ) {
      /* they are the same, replacement */
      newEntry->next = current->next;
   }
   else {
      /* insertion before current */
      newEntry->next = current;
   }

   if ( current == head ) {
      /* the list is either empty, we are inserting at the beginning,
       * or we are replacing the first entry
       */
      head = newEntry;
   }
   else {
      previous->next = newEntry;
   }

   if ( ( current != NULL ) &&
        ( strcmp( current->mimeType,
                  newEntry->mimeType ) == 0 ) ) {
      freeEntryStruct( current );
   }

   exitFunction( "addOrReplaceEntry" );

   return( head );
}  /* addOrReplaceEntry */


/**********************************************************************/

void
readEntries( char* pluginPath )
{
   int numberOfPluginPaths = countNumberOfTokens( pluginPath, ":" );
   char** pluginPathArray = NULL;
   char* pluginPathCopy = NULL;
   char* nextPluginPath = NULL;
   int i = 0;

   enterFunction( "readEntries" );

   entryHead = NULL;

   /* If there are no plugin paths, we obviously can do anything */
   if ( numberOfPluginPaths <= 0 ) {
      printError( "The pluginPath is empty." );
      return;
   }

   /* copy the pluginPath */
   pluginPathCopy = (char*)
         malloc( sizeof( char ) * ( strlen( pluginPath ) + 1 ) );
   strcpy( pluginPathCopy,
           pluginPath );

   /* allocate space for all the plug in paths */
   pluginPathArray = (char**)
         malloc( sizeof( char* ) * ( numberOfPluginPaths + 1 ) );

   /* NULL terminate the pluginPathArray */
   pluginPathArray[numberOfPluginPaths] = NULL;

   /* start tokenizing the plugin path */
   nextPluginPath = strtok( pluginPathCopy, ":" );
   
   /* put the stuff in the pluginPathArray */
   while ( nextPluginPath != NULL ) {
      pluginPathArray[i] = (char*)
            malloc( sizeof( char ) * ( strlen( nextPluginPath ) + 1 ) );
      strcpy( pluginPathArray[i],
              nextPluginPath );

      /* get the next token */
      nextPluginPath = strtok( NULL, ":" );

      /* increment the array index */
      ++i;
   }


   /* free up the memory used by pluginPathCopy */
   free( pluginPathCopy );

   /* restart the index a 0 */
   i = 0;

   /* loop through each path */
   while( pluginPathArray[i] != NULL ) {
      FILE* fptr;
      char confFileName[MAX_FILE_NAME_LENGTH];

      /* copy the plugin path */
      strcpy( confFileName,
              pluginPathArray[i] );

      /* if there is no trailing slash ('/'), add one */
      if ( confFileName[strlen( confFileName ) - 1] != '/' ) {
         strcat( confFileName,
                 "/" );
      }

      /* add the name of the configuration file to the end */
      strcat( confFileName,
              "../" );
      strcat( confFileName,
              CONF_FILE );

      printDebug( "Reading configuration from %s\n",
                  confFileName );

      /* open the file for reading */
      if ( ( fptr = fopen( confFileName, "r" ) ) == NULL ) {
         if ( errno != ENOENT ) {
            printError( "Unable to open %s for reading: %s\n",
                        confFileName,
                        strerror( errno ) );
         }
      }
      else {
         char buff[MAX_ENTRY_LENGTH];

         while ( getNextEntry( fptr, buff ) != NULL ) {
            entryStruct* newEntry = NULL;

            if ( ( newEntry = splitFields( buff ) ) == NULL ) {
               printError( "Bad line in configuration file\n\t'%s'\n",
                           buff );
            }
            else {
               /* Be sure the next link is initialized */
               newEntry->next = NULL;

               entryHead = addOrReplaceEntry( entryHead,
                                              newEntry );
            }  /* if splitFields worked */
         }  /* while more entries */
      }

      /* free up the memory taken by this pluginArrayPath */
      free( pluginPathArray[i] );

      /* increment the pluginPathArray index */
      ++i;
   }

   /* free up the memory taken up by the array */
   free( pluginPathArray );

   exitFunction( "readEntries" );
}  /* readEntries */


/**********************************************************************/

void
cleanup( void )
{
   entryStruct* currentEntry = entryHead;
  
   enterFunction( "cleanup" );

   while ( currentEntry != NULL ) {
      entryStruct* nextEntry = currentEntry->next;

      freeEntryStruct( currentEntry );

      currentEntry = nextEntry;
   }  /* while currentEntry is still real */

   exitFunction( "cleanup" );
}  /* cleanup */


/**********************************************************************/

char*
NPP_GetMIMEDescription( )
{
   char* mimeDesc;
   int mimeDescSize;
   int needSize = 1;
   entryStruct* currentEntry;

   enterFunction( "NPP_GetMIMEDescription" );

   /* read the configuration files, put information in global
    * variable entryHead
    */
   readEntries( getPluginPath() );

   /* set the current entry to the new entryHead set in readEntries */
   currentEntry = entryHead;

   /* start with one chunk of memory */
   mimeDescSize = MEMORY_CHUNK_SIZE;
   mimeDesc = (char*) malloc( sizeof( char ) * MEMORY_CHUNK_SIZE );
   mimeDesc[0] = '\0';

   /* Copy the entry array into mimeDesc */
   while ( currentEntry != NULL ) {
      char extensions[MAX_FIELD_LENGTH];
      char* description;

      /* copy the extensions */
      strcpy( extensions,
              currentEntry->extensions );

      /* increment the amount of space we need to store another entry
       * magic number of 3 == 2 seperating colons + seperating semi-colon
       */
      needSize += 3;

      /* increment the amount of space we need to store the MIME type */
      needSize += strlen( currentEntry->mimeType );

      /* add a period to each extension and increment the amount of space
       * we need to store all the information
       */
      needSize += strlen( addPeriodToEachExtension( extensions ) );
      
      /* increment the amount of space we need to store the description */
      if ( currentEntry->description != NULL ) {
         description = currentEntry->description;
      }
      else {
         description = NAME;
      }

      needSize += strlen( description );

      /* allocate space in mimeDesc as we need it */
      if ( mimeDescSize < needSize ) {
         /* store the old one so we don't lose it */
         char* originalMIMEDesc = mimeDesc;

         /* allocate another chunk of memory */
         mimeDescSize += MEMORY_CHUNK_SIZE;
         mimeDesc = (char*) malloc( sizeof( char ) * MEMORY_CHUNK_SIZE );

         /* copy the old one to the new memory space */
         strcpy( mimeDesc,
                 originalMIMEDesc );

         /* free up the memory used by the old one */
         free( originalMIMEDesc );
      }  /* if we need more memory */

      sprintf( mimeDesc,
               "%s%s:%s:%s;",
               mimeDesc,
               currentEntry->mimeType,
               extensions,
               description );

      /* move onto the next one */
      currentEntry = currentEntry->next;
   }

   /* remove the very last semi-colon */
   {
      int mimeDescLen = strlen( mimeDesc );

      if ( mimeDescLen > 0 ) {
         mimeDesc[mimeDescLen - 1] = '\0';
      }
   }

   exitFunction( "NPP_GetMIMEDescription" );

   return( mimeDesc );
}  /* NPP_GetMIMEDescription */


/**********************************************************************/

char*
findAttributeValue( int16 count,
                    char** names,
                    char** values,
                    char* name )
{
   int i;

   enterFunction( "findAttributeValue" );

   for( i = 0;
        i < count;
        i++ ) {
      if ( strcasecmp( names[i],
                       name ) == 0 ) {
         /* i is the correct attribute */
         exitFunction( "findAttributeValue" );

         return( values[i] );
      }
   }

   exitFunction( "findAttributeValue" );

   return( NULL );
}  /* findAttributeValue */


/**********************************************************************/

int
validMIMEType( entryStruct* entry,
               char* mimeType )
{
   enterFunction( "validMIMEType" );
   exitFunction( "validMIMEType" );

   return( strcmp( entry->mimeType,
                   mimeType ) );
}  /* validMIMEType */


/**********************************************************************/

entryStruct*
findEntryForMIMEType( char* mimeType )
{
   entryStruct* currentEntry = entryHead;

   enterFunction( "findEntryForMIMEType" );

   while ( currentEntry != NULL ) {
      if ( validMIMEType( currentEntry,
                          mimeType ) == 0 ) {
         exitFunction( "findEntryForMIMEType" );

         return( currentEntry );
      }

      currentEntry = currentEntry->next;
   }

   exitFunction( "findEntryForMIMEType" );

   return( NULL );
}  /* findEntryForMIMEType */


/**********************************************************************/

int
validExtension( entryStruct* entry,
                char* extension )
{
   char* extensionsCopy;
   char* nextExtension;
   int rc = 1;     /* Initialize it to not found */

   enterFunction( "validExtension" );

   /* copy the extensions */
   extensionsCopy = (char*)
         malloc( sizeof( char ) * ( strlen( entry->extensions ) + 1 ) );
   strcpy( extensionsCopy,
           entry->extensions );

   nextExtension = (char*) strtok( extensionsCopy, ", " );

   while ( ( nextExtension != NULL ) &&
           ( ( rc = strcmp( nextExtension,
                            extension ) ) != 0 ) ) {
      nextExtension = (char*) strtok( NULL, ", " );
   }

   /* free up the memory */
   free( extensionsCopy );

   exitFunction( "validExtension" );

   return( rc );
}  /* validExtension */


/**********************************************************************/

entryStruct*
findEntryForExtension( char* extension )
{
   entryStruct* currentEntry = entryHead;

   enterFunction( "findEntryForExtension" );

   while ( currentEntry != NULL ) {
      if ( validExtension( currentEntry,
                           extension ) == 0 ) {
         exitFunction( "findEntryForExtension" );

         return( currentEntry );
      }

      currentEntry = currentEntry->next;
   }

   exitFunction( "findEntryForExtension" );

   return( NULL );
}  /* findEntryForExtension */


/**********************************************************************/

NPError
NPP_GetValue( void* future,
              NPPVariable variable,
              void* value )
{
   NPError rc = NPERR_NO_ERROR;

   enterFunction( "NPP_GetValue" );

   switch( variable ) {
      case NPPVpluginNameString:
         *( (char **) value ) = NAME;
         break;
      case NPPVpluginDescriptionString:
         *( (char **) value ) = PURPOSE;				
         break;
      default:
         rc = NPERR_GENERIC_ERROR;
   }

   exitFunction( "NPP_GetValue" );

   return( rc );
}  /* NPP_GetValue */


/**********************************************************************/

NPError
NPP_Initialize( )
{
   /* Redirect standard out and err to files */
   freopen( STDERR,
            "a",
            stderr );
   freopen( STDOUT,
            "a",
            stdout );

   printMessage( "--------------------------------------------------------\n" );

   enterFunction( "NPP_Initialize" );

   readEntries( getPluginPath() );

   exitFunction( "NPP_Initialize" );

   return( NPERR_NO_ERROR );
}  /* NPP_Initialize */


/**********************************************************************/

jref
NPP_GetJavaClass( )
{
   enterFunction( "NPP_GetJavaClass" );
   exitFunction( "NPP_GetJavaClass" );

   return( NULL );
}  /* NPP_GetJavaClass */


/**********************************************************************/

void
NPP_Shutdown( )
{
   enterFunction( "NPP_Shutdown" );

   cleanup();

   exitFunction( "NPP_Shutdown" );

   printMessage( "--------------------------------------------------------\n" );
}  /* NPP_Shutdown */


/**********************************************************************/

NPError
NPP_New( NPMIMEType mimeType,
         NPP instance,
         uint16 mode,
         int16 attributeCount,
         char** attributeNames,
         char** attributeValues,
         NPSavedData* saved )
{
   NPError rc = NPERR_FILE_NOT_FOUND;

   enterFunction( "NPP_New" );

   if ( instance == NULL ) {
      rc = NPERR_INVALID_INSTANCE_ERROR;
   }
   else {
      PluginInstance* This;

      instance->pdata = NPN_MemAlloc( sizeof( PluginInstance ) );
	
      This = (PluginInstance*) instance->pdata;
      This->childPID = (pid_t) 0;
  
      /* If we know how to handle that MIME type */
      if ( ( This->entry = findEntryForMIMEType( mimeType ) ) != NULL ) {
         rc = NPERR_NO_ERROR;
      }
      else {
         /* This should never get hit, but in case it does, let's
          * try to do something intelligent
          */
         char* lastPeriod = NULL;
         char* temp;

         /* Look for the SRC attribute, it contains the source file name */
         temp = findAttributeValue( attributeCount,
                                    attributeNames,
                                    attributeValues,
                                    "SRC" );

         /* only get the extension (without the leading period) */
         if ( ( lastPeriod = (char*) strrchr( temp, '.' ) ) != NULL ) {
            strcpy( temp,
                    lastPeriod + sizeof( char ) );

            /* If the SRC attribute exists and we know how to handle
             * that type of source file
             */
            if ( ( temp != NULL ) &&
                 ( ( This->entry = findEntryForExtension( temp ) ) != NULL ) ) {
               rc = NPERR_NO_ERROR;
            }
         }
      }

      printDebug( "Found an entry that matches:\n" );
      dumpEntryStruct( This->entry );
   }

   exitFunction( "NPP_New" );

   return( rc );
}  /* NPP_New */


/**********************************************************************/

NPError 
NPP_Destroy( NPP instance,
             NPSavedData** save )
{
   NPError rc = NPERR_GENERIC_ERROR;

   enterFunction( "NPP_Destroy" );

   if ( instance == NULL ) {
      rc = NPERR_INVALID_INSTANCE_ERROR;
   }
   else {
      PluginInstance* This;

      This = (PluginInstance*) instance->pdata;

      if ( (int) This->childPID != 0 ) {
         /* kill the child associated with this instance */
         printMessage( "Killing child PID %d\n",
                       This->childPID );

         kill( This->childPID,
               SIGTERM );
      }

      /* free up the instance memory */
      NPN_MemFree( instance->pdata );

      instance->pdata = NULL;

      rc = NPERR_NO_ERROR;
   }

   exitFunction( "NPP_Destroy" );

   return( rc );
}  /* NPP_Destroy */


/**********************************************************************/

NPError 
NPP_SetWindow( NPP instance,
               NPWindow* window )
{
   NPError rc = NPERR_GENERIC_ERROR;

   enterFunction( "NPP_SetWindow" );

   if ( instance == NULL ) {
      rc = NPERR_INVALID_INSTANCE_ERROR;
   }
   else if ( window == NULL ) {
      rc = NPERR_NO_ERROR;
   }
   else {
      /* This plug-in does not yet know anything about windows, save CPU
       * cycles by not doing extraneous work
      PluginInstance* This = (PluginInstance*) instance->pdata;
      */

      rc = NPERR_NO_ERROR;
   }

   exitFunction( "NPP_SetWindow" );

   return( rc );
}  /* NPP_SetWindow */


/**********************************************************************/

NPError 
NPP_NewStream( NPP instance,
               NPMIMEType type,
               NPStream *stream, 
               NPBool seekable,
               uint16 *stype )
{
   NPError rc = NPERR_GENERIC_ERROR;

   enterFunction( "NPP_NewStream" );

   if ( instance == NULL ) {
      rc = NPERR_INVALID_INSTANCE_ERROR;
   }
   else {
      PluginInstance* This = (PluginInstance*) instance->pdata;

      /* we don't understand how to handle streams, so we must insist
       * that we have a local file name we can use instead, i.e., have
       * Netscape download the data into a temporary file and tell us
       * what the temporary file name is by calling NPP_StreamAsFile */
      *stype = NP_ASFILEONLY;

      /* initialize the child */
      This->childPID = (pid_t) 0;

      printMessage( "New stream of type '%s'\n",
                    (char*) type );

      rc = NPERR_NO_ERROR;
   }

   exitFunction( "NPP_NewStream" );

   return( rc );
}  /* NPP_NewStream */


/**********************************************************************/

int32 
NPP_WriteReady( NPP instance,
                NPStream *stream )
{
   enterFunction( "NPP_WriteReady" );

   if ( instance != NULL ) {
/*
      PluginInstance* This = (PluginInstance*) instance->pdata;
*/
   }

   exitFunction( "NPP_WriteReady" );

   /* since we don't know how to handle streams, we are sending
    * "some large number" as it tells us to do above
    */
   return( STREAMBUFSIZE );
}  /* NPP_WriteReady */


/**********************************************************************/

int32 
NPP_Write( NPP instance,
           NPStream *stream,
           int32 offset,
           int32 len,
           void *buffer )
{
   enterFunction( "NPP_Write" );

   if ( instance != NULL ) {
/*
      PluginInstance* This = (PluginInstance*) instance->pdata;
*/
   }

   exitFunction( "NPP_Write" );

   /* Make it look like we handled all the data it fed us */
   return( len );
}  /* NPP_Write */


/**********************************************************************/

NPError
NPP_DestroyStream( NPP instance,
                   NPStream *stream,
                   NPError reason )
{
   NPError rc = NPERR_GENERIC_ERROR;

   enterFunction( "NPP_DestroyStream" );

   if ( instance == NULL ) {
      rc = NPERR_INVALID_INSTANCE_ERROR;
   }
   else {
      PluginInstance* This = (PluginInstance*) instance->pdata;
  
      /* if the user clicked break, or there was 
       * a problem close down the program
       */
  
      if ( reason == NPRES_DONE ) {
         printMessage( "Stream closed, successful transfer\n" );
      }
      else if ( ( reason == NPRES_USER_BREAK ) ||
                ( reason == NPRES_NETWORK_ERR ) ) {
         printMessage( "Stream destroyed, user terminated or network error\n" );

         if ( (int) This->childPID != 0 ) {
            printMessage( "Killing child PID %d\n",
                          This->childPID );

            kill( This->childPID,
                  SIGINT );
         }
      }

      rc = NPERR_NO_ERROR;
   }
  
   exitFunction( "NPP_DestroyStream" );

   return( rc );
}  /* NPP_DestroyStream */


/**********************************************************************/

void
NPP_StreamAsFile( NPP instance,
                  NPStream *stream,
                  const char* fname )
{
   enterFunction( "NPP_StreamAsFile" );

   if ( fname != NULL ) {
      if ( instance != NULL ) {
         PluginInstance* This = (PluginInstance*) instance->pdata;
         pid_t childPID;
 
         childPID = fork();

         /* if fork failed */
         if ( childPID == -1 ) {
            printError( "Fork failed: %s\n",
                        strerror( errno ) );

            This->childPID = 0;
         }
         /* if we are the child process */
         else if ( childPID == 0 ) {
            entryStruct* entry;
            char** commandLineArray;

            printDebug( "Child process\n" );

            /* get the entry found in NPP_New */
            entry = This->entry;

            dumpEntryStruct( entry );

            commandLineArray = splitCommandLine( entry->commandLine,
                                                 (char*) fname );

            {
               char command[MAX_FILE_NAME_LENGTH];
               int i = 0;

               command[0] = '\0';

               while ( commandLineArray[i] != NULL ) {
                  strcat( command,
                          commandLineArray[i] );
                  strcat( command,
                          " " );

                  ++i;
               }

               /* Remove last space */
               command[strlen( command ) - 1] = '\0';

               printMessage( "Executing command: '%s'\n",
                              command );
            }

            /* replace this child with the application */
            execvp( commandLineArray[0],
                    commandLineArray );

            /* if we've gotten this far, the exec didn't work */
            printError( "Could not execute %s: %s\n",
                        commandLineArray[0],
                        strerror( errno ) );

            /* Exit child with a bad exit code */
            _exit( 1 );
         }
         /* if we are the parent process */
         else {
            /* save the pid so we can kill it later */
            This->childPID = childPID;

            printMessage( "Parent w/ child PID of %d\n",
                          This->childPID );
         }
      }
   }
   else {
      printError( "Error, file name is NULL\n" );
   }

   exitFunction( "NPP_StreamAsFile" );
}  /* NPP_StreamAsFile */


/**********************************************************************/

void 
NPP_Print( NPP instance,
           NPPrint* printInfo )
{
   enterFunction( "NPP_Print" );

   if ( printInfo == NULL ) {
      return;
   }

   exitFunction( "NPP_Print" );
}  /* NPP_Print */


/**********************************************************************/

void
dumpEntryStruct( entryStruct* entry )
{
#ifdef DEBUG
   enterFunction( "dumpEntryStruct" );

   if ( entry == NULL ) {
      printError( "NULL entry, not dumping\n" );
   }
   else {
      printDebug( "MIME Type: '%s'\nExtensions: '%s'\nCommand Line: '%s'\n",
                  entry->mimeType,
                  entry->extensions,
                  entry->commandLine );

      if ( entry->description != NULL ) {
         printDebug( "Description: '%s'\n",
                     entry->description );
      }
   }

   exitFunction( "dumpEntryStruct" );
#endif
}  /* dumpEntryStruct */


/**********************************************************************/

void
printMessage( char* fmt,
              ... )
{
   va_list args;

   va_start( args, fmt );

   vprintf( fmt, args );

   va_end( args );

   fflush( stdout );
}  /* printMessage */


/**********************************************************************/

void
printDebug( char* fmt,
            ... )
{
#ifdef DEBUG
   va_list args;

   va_start( args, fmt );

   vprintf( fmt, args );

   va_end( args );

   fflush( stdout );
#endif
}  /* printDebug */


/**********************************************************************/

void
printError( char* fmt,
            ... )
{
   va_list args;

   va_start( args, fmt );

   vfprintf( stderr,
             fmt,
             args );

   va_end( args );

   fflush( stderr );
}  /* printError */


/**********************************************************************/

void
enterFunction( char* functionName )
{
   printDebug( "Entering %s\n",
               functionName );
}  /* enterFunction */


/**********************************************************************/

void
exitFunction( char* functionName )
{
   printDebug( "Exiting %s\n",
               functionName );
}  /* exitFunction */
