/************************************************************

mimeplugin.h
started December 7, 1996

harshaw@wpi.edu (until may 1997)

take a look at the netscape homepage and download the 
netscape plugin SDK for a complete definition of all
API functions.

************************************************************
Copyright (C) 1996 Carl Shimer

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        

************************************************************

See CHANGES to see a complete list of changes

************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include "npapi.h"

/***********************************************************/

#define NAME "Generic Linux Plug-in"
#define PURPOSE "This plug-in can handle any embedded object with an application as specified in the configuration file."

#ifndef NS_DIR
	#define NS_DIR "/usr/local/lib/netscape/"
#endif

#ifndef CONF_FILE
   #define CONF_FILE "mimeplugin.conf"
#endif

#define MAX_ENTRY_LENGTH 512
#define MAX_FIELD_LENGTH 128
#define MAX_FILE_NAME_LENGTH 512
#define MEMORY_CHUNK_SIZE 1024


/* If we are reading from a file in NPAsFile
 * mode so we can take any size stream in our
 * write call (since we ignore it) */

int32 STREAMBUFSIZE = 0x0FFFFFFF;


/************************************************************
this struct holds all the data for a particular
entry in the configuration file

commandLine is the command line to pass to execv
mimeType is the MIME type to register with Netscape
extensions is a comma delimited list of extensions for
   the mimeType
next is a link to the next entry in the list

************************************************************/

typedef struct singleEntryStructure {
   char* mimeType;
   char* extensions;
   char* commandLine;
   char* description;
   struct singleEntryStructure* next;
} entryStruct;


/************************************************************
This is a pointer to the first entry in the list
************************************************************/

entryStruct* entryHead;


/***********************************************************************
 * Instance state information about the plugin.
 *
 * PLUGIN DEVELOPERS:
 *	Use this struct to hold per-instance information that you'll
 *	need in the various functions in this file.
 ***********************************************************************/

typedef struct _PluginInstance
{
   pid_t childPID;
   entryStruct* entry;
} PluginInstance;


/**********************************************************************
char*
getPluginPath( )

This function returns the plugin path.  If NPX_PLUGIN_PATH exists
in the environment, that is returned, else the default plugin path
is returned (See the Netscape Plugin README documentation for
details)

If NULL is returned, there is a serious problem.

**********************************************************************/

char*
getPluginPath( );


/**********************************************************************
char*
getNextEntry( FILE* fptr,
              char* buff )

This function read lines from the configuration file (fptr) until it
finds an entry.  The entry (a semi-colon seperated string) will be in
buff.  It is assumed buff is large enough to hold the next entry.

If NULL is returned, buff does NOT contain an entry.

**********************************************************************/

char*
getNextEntry( FILE* fptr,
              char* buff );


/**********************************************************************
char*
removeLeadingSpaces( char* buff )

This function removes the leading spaces from buff.

Returns the new buff without leading spaces.

**********************************************************************/

char*
removeLeadingSpaces( char* buff );


/**********************************************************************
char*
removeTrailingSpaces( char* buff )

This function removes the trailing spaces from buff.

Returns the new buff without trailing spaces.

**********************************************************************/

char*
removeTrailingSpaces( char* buff );


/**********************************************************************
char*
removeLeadingAndTrailingSpaces( char* buff )

This function removes the leading and trailing spaces from buff.

**********************************************************************/

char*
removeLeadingAndTrailingSpaces( char* buff );


/**********************************************************************
char*
removeAllSpaces( char* buff )

This function removes all spaces from buff.

**********************************************************************/

char*
removeAllSpaces( char* buff );


/**********************************************************************
void
addPeriodToEachExtension( char* listOfExtensions )

This function adds a period ('.') before each extension and removes all
spaces from the list.  It is assumed that listOfExtensions is a
comma (',') delimeted list of extensions.  It is also assumed that buff
is large enough to hold the new buff with the periods.

Returns new extensions with period before each extension.

**********************************************************************/

char*
addPeriodToEachExtension( char* extensions );


/**********************************************************************
entryStruct*
splitFields( char* buff )

This function takes a line (buff) and splits it at semi-colons (';')
and fills in the information into the structure (entry) after
allocating it.

Returns the address of the new entry structure, NULL otherwise.

Keep in mind, if the MIME type could not be obtained, it doesn't try
to get anything else.

**********************************************************************/

entryStruct*
splitFields( char* buff );


/**********************************************************************
int
countNumberOfTokens( char* str,
                     char* delimeters )

This function counts the number of tokens seperated by one of
the delimeters (delimeters) in the string (str).

Returns the number of tokens in the string.

**********************************************************************/

int
countNumberOfTokens( char* str,
                     char* delimeters );


/**********************************************************************
char**
splitCommandLine( char* commandLine,
                  char* fileName )

This function splits the command line (commandLine) on spaces and
puts each token (argument) in a command line array.  If the argument
is '%s', the file name (fileName) is put in its place instead.

Returns a pointer to the command line array.

**********************************************************************/

char**
splitCommandLine( char* commandLine,
                  char* fileName );


/**********************************************************************
void
freeEntryStruct( entryStruct* entry )

This function frees the memory used by an entry structure (entry).

**********************************************************************/

void
freeEntryStruct( entryStruct* entry );


/**********************************************************************
entryStruct*
addOrReplaceEntry( entryStruct* head,
                   entryStruct* newEntry )

This function inserts an entry into the entry list (head) if there
is not a node in the list that containing the MIME type in the new
entry (newEntry->mimeType).  If there is already node in the list
with the same MIME type, it is replaced with the new entry.

Returns a pointer to the first node in the list.

**********************************************************************/

entryStruct*
addOrReplaceEntry( entryStruct* head,
                   entryStruct* newEntry );


/**********************************************************************
void
readEntries( char* pluginPath )

this function parses the configuration file into the appropriate structures

**********************************************************************/

void
readEntries( char* pluginPath );


/**********************************************************************
void
cleanup( void )

This function cleans up all the memory we may have allocated.

**********************************************************************/

void
cleanup( void );


/**********************************************************************
char*
findAttributeValue( int16 count,
                    char** names,
                    char** values,
                    char* name )

This function finds the attribute value associated with a certain
attribute name (name).

Returns the value associated with the attribute name.

**********************************************************************/

char*
findAttributeValue( int16 count,
                    char** names,
                    char** values,
                    char* name );


/**********************************************************************
int
validMIMEType( entryStruct* entry,
               char* mimeType )

This function checks to see if a certain MIME type (mimeType)
is valid for a certain entry (entry).

Returns 0 if mimeType is a valid MIME type for the entry
entry, non-zero otherwise.

**********************************************************************/

int
validMIMEType( entryStruct* entry,
               char* mimeType );


/**********************************************************************
entryStruct*
findEntryForMIMEType( char* mimeType )

This function finds the entry associated with a certain
MIME type (mimeType).

Returns the entry associated with the MIME type.

**********************************************************************/

entryStruct*
findEntryForMIMEType( char* mimeType );


/**********************************************************************
int
validExtension( entryStruct* entry,
                char* extension )

This function checks to see if a certain extension (extension)
is valid for a certain entry (entry).

Returns 0 if extension is a valid extension for the entry
entry, non-zero otherwise.

**********************************************************************/

int
validExtension( entryStruct* entry,
                char* extension );


/**********************************************************************
entryStruct*
findEntryForExtension( char* extension )

This function finds the entry associated with a certain
extension (extension).  It is assumed that extension does not
begin with a period ('.').

Returns the entry associated with the extension.

**********************************************************************/

entryStruct*
findEntryForExtension( char* extension );


/**********************************************************************
void
dumpEntryStruct( entryStruct* entry )

This function dumps the contents of an entry structure (entry)
to stdout.

**********************************************************************/

void
dumpEntryStruct( entryStruct* entry );


/**********************************************************************
void
printMessage( char* fmt,
              ... )

This function prints a message (fmt) to stdout.  Allows for multiple
arguments if fmt is a format on how to handle ...

**********************************************************************/

void
printMessage( char* fmt,
              ... );


/**********************************************************************
void
printDebug( char* fmt,
            ... )

This function prints a message (fmt) to stdout if DEBUG is defined.
Allows for multiple arguments if fmt is a format on how to handle ...

**********************************************************************/

void
printDebug( char* fmt,
            ... );


/**********************************************************************
void
printError( char* fmt,
            ... )

This function prints a message (fmt) to stderr.  Allows for multiple
arguments if fmt is a format on how to handle ...

**********************************************************************/

void
printError( char* fmt,
            ... );


/**********************************************************************
void
enterFunction( char* functionName )

This function prints a message ("Entering " + functionName)
to stdout if DEBUG is defined.

**********************************************************************/

void
enterFunction( char* functionName );


/**********************************************************************
void
exitFunction( char* functionName )

This function prints a message ("Exiting " + functionName)
to stdout if DEBUG is defined.

**********************************************************************/

void
exitFunction( char* functionName );
